use std::cell::RefCell;

use enum_map::EnumMap;
use fluent::fluent_args;
use seed::{
    a, attrs, button, div, h1, img, input, log, prelude::*, section, span, table, td, th, tr, Url,
    C,
};

use shared::{
    apirequests::general::Ordering,
    apirequests::{
        general::{EditMode, Message, Operation, Status},
        links::{LinkDelta, LinkOverviewColumns, LinkRequestForm},
    },
    datatypes::FullLink,
};

use crate::{i18n::I18n, unwrap_or_return, unwrap_or_send};


/// Setup the page
pub fn init(mut url: Url, orders: &mut impl Orders<Msg>, i18n: I18n) -> Model {
    // fetch the links to fill the list.
    orders.send_msg(Msg::Query(QueryMsg::Fetch));
    // if the url contains create_link set the edit_link variable.
    // This variable then opens the create link dialog.
    let edit_link = match url.next_path_part() {
        Some("create_link") => Some(RefCell::new(LinkDelta::default())),
        None | Some(_) => None,
    };
    
    Model {
        links: Vec::new(), // will contain the links to display
        i18n, // to translate
        formconfig: LinkRequestForm::default(), // when requesting links the form is stored here
        inputs: EnumMap::default(), // the input fields for the searches
        edit_link, // if set this will trigger a link edit dialog
        last_message: None, // if a message to the user is recieved
        question: None, // some operations should be confirmed
    }
}

#[derive(Debug)]
pub struct Model {
    links: Vec<FullLink>, // will contain the links to display
    i18n: I18n, // to translate
    formconfig: LinkRequestForm, // when requesting links the form is stored here
    inputs: EnumMap<LinkOverviewColumns, FilterInput>, // the input fields for the searches
    edit_link: Option<RefCell<LinkDelta>>, // if set this will trigger a link edit dialog
    last_message: Option<Status>, // if a message to the user is recieved
    question: Option<EditMsg>, // some operations should be confirmed
}

#[derive(Default, Debug, Clone)]
struct FilterInput {
    filter_input: ElRef<web_sys::HtmlInputElement>,
}

#[derive(Clone)]
pub enum Msg {
    Query(QueryMsg), // Messages related to querying links
    Edit(EditMsg), // Messages related to editing links
    ClearAll, // Clear all messages
    SetMessage(String), // Set a message to the user
}

#[derive(Clone)]
pub enum QueryMsg {
    Fetch,
    OrderBy(LinkOverviewColumns),
    Received(Vec<FullLink>),
    CodeFilterChanged(String),
    DescriptionFilterChanged(String),
    TargetFilterChanged(String),
    AuthorFilterChanged(String),
}
/// All the messages on link editing
#[derive(Clone, Debug)]
pub enum EditMsg {
    EditSelected(LinkDelta),
    CreateNewLink,
    Created(Status),
    EditCodeChanged(String),
    EditDescriptionChanged(String),
    EditTargetChanged(String),
    MayDeleteSelected(LinkDelta),
    DeleteSelected(LinkDelta),
    SaveLink,
    FailedToCreateLink,
    FailedToDeleteLink,
    DeletedLink(Status),
}

/// hide all dialogs
fn clear_all(model: &mut Model) {
    model.edit_link = None;
    model.last_message = None;
    model.question = None;
}

/// React to environment changes
pub fn update(msg: Msg, model: &mut Model, orders: &mut impl Orders<Msg>) {
    match msg {
        Msg::Query(msg) => process_query_messages(msg, model, orders),
        Msg::Edit(msg) => process_edit_messages(msg, model, orders),
        Msg::ClearAll => {
            clear_all(model)
        }
        Msg::SetMessage(msg) => {
            clear_all(model);
            model.last_message = Some(Status::Error(Message { message: msg }));
        }
    }
}

/// Process all messages for loading the information from the server.
pub fn process_query_messages(msg: QueryMsg, model: &mut Model, orders: &mut impl Orders<Msg>) {
    match msg {
        QueryMsg::Fetch => {
            orders.skip(); // No need to rerender
            load_links(model, orders)
        }
        // Default to ascending ordering but if the links are already sorted according to this collumn toggle between ascending and descending ordering.
        QueryMsg::OrderBy(column) => {
            model.formconfig.order = model.formconfig.order.as_ref().map_or_else(
                || {
                    Some(Operation {
                        column: column.clone(),
                        value: Ordering::Ascending,
                    })
                },
                |order| {
                    Some(Operation {
                        column: column.clone(),
                        value: if order.column == column && order.value == Ordering::Ascending {
                            Ordering::Descending
                        } else {
                            Ordering::Ascending
                        },
                    })
                },
            );
            // After setting up the ordering fetch the links from the server again with the new filter settings.
            // If the new filters and ordering include more links the list would be incomplete otherwise.
            orders.send_msg(Msg::Query(QueryMsg::Fetch));

            // Also sort the links locally - can probably removed...
            model.links.sort_by(match column {
                LinkOverviewColumns::Code => {
                    |o: &FullLink, t: &FullLink| o.link.code.cmp(&t.link.code)
                }
                LinkOverviewColumns::Description => {
                    |o: &FullLink, t: &FullLink| o.link.title.cmp(&t.link.title)
                }
                LinkOverviewColumns::Target => {
                    |o: &FullLink, t: &FullLink| o.link.target.cmp(&t.link.target)
                }
                LinkOverviewColumns::Author => {
                    |o: &FullLink, t: &FullLink| o.user.username.cmp(&t.user.username)
                }
                LinkOverviewColumns::Statistics => {
                    |o: &FullLink, t: &FullLink| o.clicks.number.cmp(&t.clicks.number)
                }
            })
        }
        QueryMsg::Received(response) => {
            model.links = response;
        }
        QueryMsg::CodeFilterChanged(s) => {
            log!("Filter is: ", &s);
            let sanit = s.chars().filter(|x| x.is_alphanumeric()).collect();
            model.formconfig.filter[LinkOverviewColumns::Code].sieve = sanit;
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
        QueryMsg::DescriptionFilterChanged(s) => {
            log!("Filter is: ", &s);
            let sanit = s.chars().filter(|x| x.is_alphanumeric()).collect();
            model.formconfig.filter[LinkOverviewColumns::Description].sieve = sanit;
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
        QueryMsg::TargetFilterChanged(s) => {
            log!("Filter is: ", &s);
            let sanit = s.chars().filter(|x| x.is_alphanumeric()).collect();
            model.formconfig.filter[LinkOverviewColumns::Target].sieve = sanit;
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
        QueryMsg::AuthorFilterChanged(s) => {
            log!("Filter is: ", &s);
            let sanit = s.chars().filter(|x| x.is_alphanumeric()).collect();
            model.formconfig.filter[LinkOverviewColumns::Author].sieve = sanit;
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
    }
}

/// Perform a request to the server to load the links from the server.
fn load_links(model: &Model, orders: &mut impl Orders<Msg>) {
    let data = model.formconfig.clone();
    orders.perform_cmd(async {
        let data = data;
        // create a request
        let request = unwrap_or_return!(
            Request::new("/admin/json/list_links/")
                .method(Method::Post)
                .json(&data),
            Msg::SetMessage("Failed to parse data".to_string())
        );
        // send the request and recieve a response
        let response = unwrap_or_return!(
            fetch(request).await,
            Msg::SetMessage("Failed to send data".to_string())
        );
        // check the html status to be 200
        let response = unwrap_or_return!(
            response.check_status(),
            Msg::SetMessage("Wrong response code".to_string())
        );
        // unpack the response into the `Vec<FullLink>`
        let links: Vec<FullLink> = unwrap_or_return!(
            response.json().await,
            Msg::SetMessage("Invalid response".to_string())
        );
        // The message that is sent by perform_cmd after this async block is completed
        Msg::Query(QueryMsg::Received(links))
    });
}

/// Process all the events related to editing links.
pub fn process_edit_messages(msg: EditMsg, model: &mut Model, orders: &mut impl Orders<Msg>) {
    match msg {
        EditMsg::EditSelected(link) => {
            clear_all(model);
            model.edit_link = Some(RefCell::new(link))
        }
        EditMsg::CreateNewLink => {
            clear_all(model);
            model.edit_link = Some(RefCell::new(LinkDelta::default()))
        }
        EditMsg::Created(success_msg) => {
            clear_all(model);
            model.last_message = Some(success_msg);
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
        EditMsg::EditCodeChanged(s) => {
            if let Some(ref le) = model.edit_link {
                le.try_borrow_mut().expect("Failed to borrow mutably").code = s;
            }
        }
        EditMsg::EditDescriptionChanged(s) => {
            if let Some(ref le) = model.edit_link {
                le.try_borrow_mut().expect("Failed to borrow mutably").title = s;
            }
        }
        EditMsg::EditTargetChanged(s) => {
            if let Some(ref le) = model.edit_link {
                le.try_borrow_mut()
                    .expect("Failed to borrow mutably")
                    .target = s;
            }
        }
        EditMsg::SaveLink => {
            save_link(model, orders);
        }
        EditMsg::FailedToCreateLink => {
            orders.send_msg(Msg::SetMessage("Failed to create this link!".to_string()));
            log!("Failed to create Link");
        }
        // capture including the message part
        link @ EditMsg::MayDeleteSelected(..) => {
            clear_all(model);
            model.question = Some(link)
        }
        EditMsg::DeleteSelected(link) => {
            orders.perform_cmd(async {
                let data = link;
                // create the request
                let request = unwrap_or_return!(
                    Request::new("/admin/json/delete_link/")
                        .method(Method::Post)
                        .json(&data),
                    Msg::SetMessage("serialization failed".to_string())
                );
                // perform the request and recieve a respnse
                let response =
                    unwrap_or_return!(fetch(request).await, Msg::Edit(EditMsg::FailedToDeleteLink));

                // check the status of the response
                let response = unwrap_or_return!(
                    response.check_status(),
                    Msg::SetMessage("Wrong response code!".to_string())
                );
                // deserialize the response
                let message: Status = unwrap_or_return!(
                    response.json().await,
                    Msg::SetMessage(
                        "Failed to parse the response! The link might or might not be deleted!"
                            .to_string()
                    )
                );

                Msg::Edit(EditMsg::DeletedLink(message))
            });
        }
        EditMsg::FailedToDeleteLink => {
            log!("Failed to delete Link");
        }
        EditMsg::DeletedLink(message) => {
            clear_all(model);
            model.last_message = Some(message);
            orders.send_msg(Msg::Query(QueryMsg::Fetch));
        }
    }
}

/// Send a link save request to the server.
fn save_link(model: &Model, orders: &mut impl Orders<Msg>) {
    // get the link to save
    let edit_link = unwrap_or_send!(
        model.edit_link.as_ref(),
        Msg::SetMessage("Please enter a link".to_string()),
        orders
    );
    let data = edit_link.borrow().clone();
    orders.perform_cmd(async {
        let data = data;
        // create the request
        let request = unwrap_or_return!(
            Request::new(match data.edit {
                EditMode::Create => "/admin/json/create_link/",
                EditMode::Edit => "/admin/json/edit_link/",
            })
            .method(Method::Post)
            .json(&data),
            Msg::SetMessage("Failed to encode the link!".to_string())
        );
        // perform the request
        let response =
            unwrap_or_return!(fetch(request).await, Msg::Edit(EditMsg::FailedToCreateLink));

        // check the response status
        let response = unwrap_or_return!(
            response.check_status(),
            Msg::SetMessage("Wrong response code".to_string())
        );
        // Parse the response
        let message: Status = unwrap_or_return!(
            response.json().await,
            Msg::SetMessage("Invalid response!".to_string())
        );

        Msg::Edit(EditMsg::Created(message))
    });
}

/// view the page
///   * messages
///   * questions
///   * the table of links including sorting and searching
#[must_use]
pub fn view(model: &Model) -> Node<Msg> {
    let lang = &model.i18n.clone();
    // shortcut for translating
    let t = move |key: &str| lang.translate(key, None);
    section![
        // display a message if any
        if let Some(message) = &model.last_message {
            div![
                C!["message", "center"],
                close_button(),
                match message {
                    Status::Success(m) | Status::Error(m) => &m.message,
                }
            ]
        } else {
            section![]
        },
        // Display a question if any
        if let Some(question) = &model.question {
            div![
                C!["message", "center"],
                close_button(),
                if let EditMsg::MayDeleteSelected(l) = question.clone() {
                    div![
                        lang.translate(
                            "really-delete",
                            Some(&fluent_args!["code" => l.code.clone()])
                        ),
                        a![t("no"), C!["button"], ev(Ev::Click, |_| Msg::ClearAll)],
                        a![
                            t("yes"),
                            C!["button"],
                            ev(Ev::Click, move |_| Msg::Edit(EditMsg::DeleteSelected(l)))
                        ]
                    ]
                } else {
                    span!()
                }
            ]
        } else {
            section![]
        },

        // display the list of links
        table![
            // Add the headlines
            view_link_table_head(&t),
            // Add filter fields right below the headlines
            view_link_table_filter_input(model, &t),
            // Add all the content lines
            model.links.iter().map(view_link)
        ],

        // A fetch button - this should not be needed and will be removed in future.
        button![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::Fetch)),
            "Fetch links"
        ],
        // Display the edit dialog if any
        if let Some(l) = &model.edit_link {
            edit_or_create_link(l, t)
        } else {
            section!()
        }
    ]
}

/// Create the headlines of the link table
fn view_link_table_head<F: Fn(&str) -> String>(t: F) -> Node<Msg> {
    tr![
        th![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::OrderBy(
                LinkOverviewColumns::Code
            ))),
            t("link-code")
        ],
        th![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::OrderBy(
                LinkOverviewColumns::Description
            ))),
            t("link-description")
        ],
        th![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::OrderBy(
                LinkOverviewColumns::Target
            ))),
            t("link-target")
        ],
        th![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::OrderBy(
                LinkOverviewColumns::Author
            ))),
            t("username")
        ],
        th![
            ev(Ev::Click, |_| Msg::Query(QueryMsg::OrderBy(
                LinkOverviewColumns::Statistics
            ))),
            t("statistics")
        ],
        th![]
    ]
}

/// Create the filter fields in the table columns
fn view_link_table_filter_input<F: Fn(&str) -> String>(model: &Model, t: F) -> Node<Msg> {
    tr![
        C!["filters"],
        td![input![
            attrs! {
                At::Value => &model.formconfig.filter[LinkOverviewColumns::Code].sieve,
                At::Type => "search",
                At::Placeholder => t("search-placeholder")
            },
            input_ev(Ev::Input, |s| Msg::Query(QueryMsg::CodeFilterChanged(s))),
            el_ref(&model.inputs[LinkOverviewColumns::Code].filter_input),
        ]],
        td![input![
            attrs! {At::Value =>
            &model
                .formconfig.filter[LinkOverviewColumns::Description].sieve,
                At::Type => "search",
                At::Placeholder => t("search-placeholder")
            },
            input_ev(Ev::Input, |s| Msg::Query(
                QueryMsg::DescriptionFilterChanged(s)
            )),
            el_ref(&model.inputs[LinkOverviewColumns::Description].filter_input),
        ]],
        td![input![
            attrs! {At::Value =>
            &model
                .formconfig.filter[LinkOverviewColumns::Target].sieve,
                At::Type => "search",
                At::Placeholder => t("search-placeholder")
            },
            input_ev(Ev::Input, |s| Msg::Query(QueryMsg::TargetFilterChanged(s))),
            el_ref(&model.inputs[LinkOverviewColumns::Target].filter_input),
        ]],
        td![input![
            attrs! {At::Value =>
            &model
                .formconfig.filter[LinkOverviewColumns::Author].sieve,
                At::Type => "search",
                At::Placeholder => t("search-placeholder")
            },
            input_ev(Ev::Input, |s| Msg::Query(QueryMsg::AuthorFilterChanged(s))),
            el_ref(&model.inputs[LinkOverviewColumns::Author].filter_input),
        ]],
        // statistics and the delete column cannot be filtered
        td![],
        td![],
    ]
}

/// display a single link
fn view_link(l: &FullLink) -> Node<Msg> {
    // Ugly hack - this is needed to be able to move the l into the closures... l.clone() in place does not work.
    let link = LinkDelta::from(l.clone());
    let link2 = LinkDelta::from(l.clone());
    let link3 = LinkDelta::from(l.clone());
    let link4 = LinkDelta::from(l.clone());
    let link5 = LinkDelta::from(l.clone());
    tr![
        {
            td![
                ev(Ev::Click, |_| Msg::Edit(EditMsg::EditSelected(link))),
                &l.link.code
            ]
        },
        {
            td![
                ev(Ev::Click, |_| Msg::Edit(EditMsg::EditSelected(link2))),
                &l.link.title
            ]
        },
        td![a![attrs![At::Href => &l.link.target], &l.link.target]],
        {
            td![
                ev(Ev::Click, |_| Msg::Edit(EditMsg::EditSelected(link3))),
                &l.user.username
            ]
        },
        {
            td![
                ev(Ev::Click, |_| Msg::Edit(EditMsg::EditSelected(link4))),
                &l.clicks.number
            ]
        },
        {
            td![img![
                ev(Ev::Click, |_| Msg::Edit(EditMsg::MayDeleteSelected(link5))),
                C!["trashicon"],
                attrs!(At::Src => "/static/trash.svg")
            ]]
        },
    ]
}

/// display a link editing dialog with save and close button
fn edit_or_create_link<F: Fn(&str) -> String>(l: &RefCell<LinkDelta>, t: F) -> Node<Msg> {
    let link = l.borrow();
    div![
        // close button top right
        C!["editdialog", "center"],
        close_button(),
        h1![match &link.edit {
            EditMode::Edit => t("edit-link"),
            EditMode::Create => t("create-link"),
        }],
        table![
            tr![
                th![t("link-description")],
                td![input![
                    attrs! {
                        At::Value => &link.title,
                        At::Type => "text",
                        At::Placeholder => t("link-description")
                    },
                    input_ev(Ev::Input, |s| {
                        Msg::Edit(EditMsg::EditDescriptionChanged(s))
                    }),
                ]]
            ],
            tr![
                th![t("link-target")],
                td![input![
                    attrs! {
                        At::Value => &link.target,
                        At::Type => "text",
                        At::Placeholder => t("link-target")
                    },
                    input_ev(Ev::Input, |s| { Msg::Edit(EditMsg::EditTargetChanged(s)) }),
                ]]
            ],
            tr![
                th![t("link-code")],
                td![input![
                    attrs! {
                        At::Value => &link.code,
                        At::Type => "text",
                        At::Placeholder => t("password")
                    },
                    input_ev(Ev::Input, |s| { Msg::Edit(EditMsg::EditCodeChanged(s)) }),
                ],]
            ]
        ],
        a![
            match &link.edit {
                EditMode::Edit => t("edit-link"),
                EditMode::Create => t("create-link"),
            },
            C!["button"],
            ev(Ev::Click, |_| Msg::Edit(EditMsg::SaveLink))
        ]
    ]
}


/// a close button for dialogs
fn close_button() -> Node<Msg> {
    div![
        C!["closebutton"],
        a!["\u{d7}"],
        ev(Ev::Click, |_| Msg::ClearAll)
    ]
}